package de.tudarmstadt.fegelod.generators;

import java.util.Enumeration;
import java.util.Vector;

import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Option;
import weka.core.Utils;

import com.hp.hpl.jena.query.QuerySolution;
import com.hp.hpl.jena.query.ResultSet;

import de.tudarmstadt.fegelod.util.AttributeTypeGuesser.attributeType;

/**
 * This generator creates a binary attribute (true|false) for incoming and outgoing
 * relations to individuals
 * e.g. for the triple
 * 	<http://foo.bar/s> <http://foo.bar/p> <http://foo.bar/o>
 * the relation uri_in_individual_http://foo.bar/p_http://foo.bar/o would be set to true.
 * 
 * @author paulheim
 *
 */
public class RelationIndividualFeatureGenerator extends
		AbstractLODFeatureGenerator {

	private static final long serialVersionUID = -7091662773052559301L;
	
	private FastVector attributeValues = new FastVector();
	
	private String propertyNamespace;
	private String individualNamespace;
	
	public RelationIndividualFeatureGenerator() {
		attributeValues.addElement("true");
		attributeValues.addElement("false");
	}
	
	@Override
	public Enumeration<?> listOptions() {
        Vector<Option> vector = new Vector<Option>();

        Enumeration<?> e = super.listOptions();
        while(e.hasMoreElements())
        	vector.add((Option)e.nextElement());
        
        vector.addElement(
    		new Option(
    		"\tuse only properties from a given namespace",
    		"PNS", 1, "-PNS <string>"));
        
        vector.addElement(
    		new Option(
    		"\tuse only individuals from a given namespace",
    		"INS", 1, "-INS <string>"));
        
        return vector.elements();
	}
	
	@Override
	public void setOptions(String[] options) throws Exception {
		super.setOptions(options);
		
		String pn = Utils.getOption("PNS", options);
		if(pn!=null && pn.length()>0)
			propertyNamespace = pn;

		String in = Utils.getOption("INS", options);
		if(in!=null && in.length()>0)
			individualNamespace = in;
	}

	@Override
	protected void processInstance(Instance instance) {
		String uri = instance.stringValue(attributeToExpand);
		// incoming
		ResultSet RS1 = queryRunner.runSelectQuery(getSPARQLQueryIncoming(uri));
		while(RS1.hasNext()) {
			QuerySolution sol = RS1.next();
			String attName = attributeToExpand.name() + "_in_individual_" + sol.get("p").toString() + "_" + sol.get("s").toString();
			Attribute thisAtt = instances.attribute(attName);
			if(thisAtt==null) {
				thisAtt = createAttribute(attName, attributeType.nominal, attributeValues, "false");
			}
			instance.setValue(thisAtt, "true");
		}
		// outgoing
		ResultSet RS2 = queryRunner.runSelectQuery(getSPARQLQueryOutgoing(uri));
		while(RS2.hasNext()) {
			QuerySolution sol = RS2.next();
			String attName = attributeToExpand.name() + "_out_individual_" + sol.get("p").toString() + "_" + sol.get("o").toString();
			Attribute thisAtt = instances.attribute(attName);
			if(thisAtt==null) {
				thisAtt = createAttribute(attName, attributeType.nominal, attributeValues, "false");
			}
			instance.setValue(thisAtt, "true");
		}
	}

	@Override
	public String globalInfo() {
		return "This generator creates a binary attribute (true|false) for incoming and outgoing relations to individuals.";
	}

	private String getSPARQLQueryOutgoing(String uri) {
		String filters = "";
		if(propertyNamespace!=null)
			if(filters.length()==0)
				filters += " REGEX(STR(?p),'^" + propertyNamespace + "')";
			else
				filters += " && REGEX(STR(?p),'^" + propertyNamespace + "')";
		if(individualNamespace!=null)
			if(filters.length()==0)
				filters += " REGEX(STR(?o),'^" + individualNamespace + "')";
			else
				filters += " && REGEX(STR(?o),'^" + individualNamespace + "')";
			
		return 	"SELECT DISTINCT ?p ?o " + 
				"FROM <http://dbpedia.org> " +
				"WHERE {<" + uri + "> ?p ?o . FILTER(isURI(?o) " + filters + ")}";
	}
	
	private String getSPARQLQueryIncoming(String uri) {
		String filters = "";
		if(propertyNamespace!=null)
			filters += " && REGEX(STR(?p),'^" + propertyNamespace + "')";
		if(individualNamespace!=null)
			filters += " && REGEX(STR(?s),'^" + individualNamespace + "')";

		if(filters.length()>0)
			return 	"SELECT DISTINCT ?p ?s " + 
			"FROM <http://dbpedia.org> " +
			"WHERE {?s ?p <" + uri + ">. FILTER(" + filters + ")}";
		else
			return 	"SELECT DISTINCT ?p ?s " + 
					"FROM <http://dbpedia.org> " +
					"WHERE {?s ?p <" + uri + ">.}";
	}
	
}
